// This file is part of the Polymesh distribution (https://github.com/PolymathNetwork/Polymesh).
// Copyright (c) 2020 Polymath

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3.

// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

//! # Identity module
//!
//! This module is used to manage identity concept.
//!
//!  - [Module](./struct.Module.html)
//!  - [Trait](./trait.Trait.html)
//!
//! ## Overview :
//!
//! Identity concept groups different account (keys) in one place, and it allows each key to
//! make operations based on the constraint that each account (permissions and key types).
//!
//! Any account can create and manage one and only one identity, using
//! [register_did](./struct.Module.html#method.register_did). Other accounts can be added to a
//! target identity as signing key, where we also define the type of account (`External`,
//! `MultiSign`, etc.) and/or its permission.
//!
//! Some operations at identity level are only allowed to its administrator account, like
//! [set_master_key](./struct.Module.html#method.set_master_key) or
//!
//! ## Identity information
//!
//! Identity contains the following data:
//!  - `master_key`. It is the administrator account of the identity.
//!  - `signing_keys`. List of keys and their capabilities (type of key and its permissions) .
//!
//! ## Freeze signing keys
//!
//! It is an *emergency action* to block all signing keys of an identity and it can only be performed
//! by its administrator.
//!
//! see [freeze_signing_keys](./struct.Module.html#method.freeze_signing_keys)
//! see [unfreeze_signing_keys](./struct.Module.html#method.unfreeze_signing_keys)
//!
//! ## Interface
//!
//! ### Dispatchable Functions
//!
//! - `register_did` - Register a new did with a CDD claim for the caller.
//! - `cdd_register_did` - Registers a new did for the target and attaches a CDD claim to it.
//! - `invalidate_cdd_claims` - Invalidates any claim generated by `cdd` from `disable_from` timestamps.
//! - `remove_signing_items` - Removes specified signing keys of a DID if present.
//! - `set_master_key` - Sets a new master key for a DID.
//! - `accept_master_key` - Accept authorization to become the new master key of an identity.
//! - `change_cdd_requirement_for_mk_rotation` - Sets if CDD authorization is required for updating master key of an identity.
//! - `join_identity_as_key` - Join an identity as a signing key.
//! - `join_identity_as_identity` - Join an identity as a signing identity.
//! - `add_claim` - Adds a new claim record or edits an existing one.
//! - `add_claims_batch` - Adds a new batch of claim records or edits an existing one.
//! - `forwarded_call` - Creates a call on behalf of another DID.
//! - `revoke_claim` - Marks the specified claim as revoked.
//! - `revoke_claims_batch` - Revokes multiple claims in a batch.
//! - `set_permission_to_signer` - Sets permissions for an specific `target_key` key.
//! - `freeze_signing_keys` - Disables all signing keys at `did` identity.
//! - `unfreeze_signing_keys` - Re-enables all signing keys of the caller's identity.
//! - `add_authorization` - Adds an authorization.
//! - `add_authorization_as_key` - Adds an authorization as a key.
//! - `batch_add_authorization` - Adds an array of authorizations.
//! - `remove_authorization` - Removes an authorization.
//! - `batch_remove_authorization` - Removes an array of authorizations.
//! - `accept_authorization` - Accepts an authorization.
//! - `batch_accept_authorization` - Accepts an array of authorizations.
//! - `add_signing_items_with_authorization` - Adds signing keys to target identity `id`.
//! - `revoke_offchain_authorization` - Revokes the `auth` off-chain authorization of `signer`.

#![cfg_attr(not(feature = "std"), no_std)]
#![recursion_limit = "256"]

#[cfg(feature = "runtime-benchmarks")]
pub mod benchmarking;

use pallet_identity_rpc_runtime_api::{DidRecords as RpcDidRecords, DidStatus, LinkType};
use pallet_transaction_payment::{CddAndFeeDetails, ChargeTxFee};
use polymesh_common_utilities::{
    constants::{
        did::{CDD_PROVIDERS_DID, GOVERNANCE_COMMITTEE_DID, SECURITY_TOKEN, USER},
        TREASURY_MODULE_ID,
    },
    protocol_fee::{ChargeProtocolFee, ProtocolOp},
    traits::{
        asset::AcceptTransfer,
        group::{GroupTrait, InactiveMember},
        identity::{
            AuthorizationNonce, IdentityTrait, LinkedKeyInfo, RawEvent, SigningItemWithAuth,
            TargetIdAuthorization, Trait,
        },
        multisig::AddSignerMultiSig,
    },
    Context, SystematicIssuers,
};
use polymesh_primitives::{
    AccountKey, AuthIdentifier, Authorization, AuthorizationData, AuthorizationError, Claim,
    ClaimType, Identity as DidRecord, IdentityClaim, IdentityId, Link, LinkData, Permission, Scope,
    Signatory, SignatoryType, SigningItem, Ticker,
};

use codec::{Decode, Encode};
use core::{
    convert::{From, TryInto},
    result::Result as StdResult,
};
use sp_core::sr25519::{Public, Signature};
use sp_io::hashing::blake2_256;
use sp_runtime::{
    traits::{
        AccountIdConversion, CheckedAdd, Dispatchable, Hash, SaturatedConversion, Verify, Zero,
    },
    AnySignature,
};
use sp_std::{convert::TryFrom, mem::swap, prelude::*, vec};

use frame_support::{
    decl_error, decl_module, decl_storage,
    dispatch::{DispatchError, DispatchResult},
    ensure,
    traits::{ChangeMembers, InitializeMembers},
    weights::{DispatchClass, FunctionOf, GetDispatchInfo, SimpleDispatchInfo},
    StorageDoubleMap,
};
use frame_system::{self as system, ensure_root, ensure_signed};

pub type Event<T> = polymesh_common_utilities::traits::identity::Event<T>;

#[derive(Encode, Decode, Clone, PartialEq, Eq, Debug, PartialOrd, Ord)]
pub struct Claim1stKey {
    pub target: IdentityId,
    pub claim_type: ClaimType,
}

#[derive(Encode, Decode, Clone, PartialEq, Eq, Debug, PartialOrd, Ord)]
pub struct Claim2ndKey {
    pub issuer: IdentityId,
    pub scope: Option<Scope>,
}

#[derive(Encode, Decode, Clone, PartialEq, Eq, Debug, PartialOrd, Ord, Default)]
pub struct BatchAddClaimItem<M> {
    pub target: IdentityId,
    pub claim: Claim,
    pub expiry: Option<M>,
}

#[derive(Encode, Decode, Clone, PartialEq, Eq, Debug, PartialOrd, Ord, Default)]
pub struct BatchRevokeClaimItem {
    pub target: IdentityId,
    pub claim: Claim,
}

decl_storage! {
    trait Store for Module<T: Trait> as identity {

        /// DID -> identity info
        pub DidRecords get(fn did_records) config(): map hasher(twox_64_concat) IdentityId => DidRecord;

        /// DID -> bool that indicates if signing keys are frozen.
        pub IsDidFrozen get(fn is_did_frozen): map hasher(twox_64_concat) IdentityId => bool;

        /// It stores the current identity for current transaction.
        pub CurrentDid: Option<IdentityId>;

        /// It stores the current gas fee payer for the current transaction
        pub CurrentPayer: Option<Signatory>;

        /// (Target ID, claim type) (issuer,scope) -> Associated claims
        pub Claims: double_map hasher(blake2_128_concat) Claim1stKey, hasher(blake2_128_concat) Claim2ndKey => IdentityClaim;

        // Account => DID
        pub KeyToIdentityIds get(fn key_to_identity_ids) config(): map hasher(blake2_128_concat) AccountKey => Option<LinkedKeyInfo>;

        /// Nonce to ensure unique actions. starts from 1.
        pub MultiPurposeNonce get(fn multi_purpose_nonce) build(|_| 1u64): u64;

        /// Authorization nonce per Identity. Initially is 0.
        pub OffChainAuthorizationNonce get(fn offchain_authorization_nonce): map hasher(twox_64_concat) IdentityId => AuthorizationNonce;

        /// Inmediate revoke of any off-chain authorization.
        pub RevokeOffChainAuthorization get(fn is_offchain_authorization_revoked): map hasher(blake2_128_concat) (Signatory, TargetIdAuthorization<T::Moment>) => bool;

        /// All authorizations that an identity/key has
        pub Authorizations: double_map hasher(blake2_128_concat) Signatory, hasher(twox_64_concat) u64 => Authorization<T::Moment>;

        /// All links that an identity/key has
        pub Links: double_map hasher(blake2_128_concat) Signatory, hasher(twox_64_concat) u64 => Link<T::Moment>;

        /// All authorizations that an identity/key has given. (Authorizer, auth_id -> authorized)
        pub AuthorizationsGiven: double_map hasher(blake2_128_concat) Signatory, hasher(twox_64_concat) u64 => Signatory;

        /// It defines if authorization from a CDD provider is needed to change master key of an identity
        pub CddAuthForMasterKeyRotation get(fn cdd_auth_for_master_key_rotation): bool;
    }
    add_extra_genesis {
        config(identities): Vec<(T::AccountId, IdentityId, IdentityId, Option<u64>)>;
        config(signing_keys): Vec<(T::AccountId, IdentityId)>;
        build(|config: &GenesisConfig<T>| {
            // Add systematic CDD for the Treasury module
            let treasury_account_id: T::AccountId = TREASURY_MODULE_ID.into_account();
            let treasury_master_key = AccountKey::try_from(treasury_account_id.encode()).unwrap();
            let treasury_did = SystematicIssuers::TreasuryModule.as_id();
            <Module<T>>::link_key_to_did(&treasury_master_key, SignatoryType::External, treasury_did);
            let record = DidRecord {
                master_key: treasury_master_key,
                ..Default::default()
            };
            <DidRecords>::insert(&treasury_did, record);
            <Module<T>>::deposit_event(RawEvent::DidCreated(treasury_did.clone(), treasury_account_id.clone(), vec![]));

            // Add the claim data for the CustomerDueDiligence type claim for Treasury module
            let claim_type = ClaimType::CustomerDueDiligence;
            let pk = Claim1stKey{ target: treasury_did, claim_type };
            let sk = Claim2ndKey{ issuer: treasury_did, scope: None };
            let id_claim = IdentityClaim {
                claim_issuer: treasury_did,
                issuance_date: 0_u64,
                last_update_date: 0_u64,
                expiry: None,
                claim: Claim::CustomerDueDiligence,
            };

            <Claims>::insert(&pk, &sk, id_claim.clone());
            <Module<T>>::deposit_event(RawEvent::ClaimAdded(treasury_did, id_claim));

            // Add System DID: Governance committee && CDD providers
            [GOVERNANCE_COMMITTEE_DID, CDD_PROVIDERS_DID, ].iter()
                .for_each(|raw_id| {
                    let id = IdentityId::from(**raw_id);
                    let master_key = AccountKey::from(**raw_id);

                    <DidRecords>::insert( id, DidRecord {
                        master_key,
                        ..Default::default()
                    });
                    <Module<T>>::deposit_event(RawEvent::DidCreated(id.clone(), T::AccountId::decode(&mut master_key.as_slice()).unwrap(), vec![]));
                });

            //  Other
            for &(ref master_account_id, issuer, did, expiry) in &config.identities {
                // Direct storage change for registering the DID and providing the claim
                let master_key = AccountKey::try_from(master_account_id.encode()).unwrap();
                assert!(!<DidRecords>::contains_key(did), "Identity already exist");
                <MultiPurposeNonce>::mutate(|n| *n += 1_u64);
                <Module<T>>::link_key_to_did(&master_key, SignatoryType::External, did);
                let record = DidRecord {
                    master_key,
                    ..Default::default()
                };
                <DidRecords>::insert(&did, record);
                <Module<T>>::deposit_event(RawEvent::DidCreated(did.clone(), master_account_id.clone(), vec![]));
                // Add the claim data for the CustomerDueDiligence type claim
                let claim_type = ClaimType::CustomerDueDiligence;
                let pk = Claim1stKey{ target: did, claim_type };
                let sk = Claim2ndKey{ issuer, scope: None };
                let id_claim = IdentityClaim {
                    claim_issuer: issuer,
                    issuance_date: 0_u64,
                    last_update_date: 0_u64,
                    expiry: expiry,
                    claim: Claim::CustomerDueDiligence,
                };

                <Claims>::insert(&pk, &sk, id_claim.clone());
                <Module<T>>::deposit_event(RawEvent::ClaimAdded(did, id_claim));
            }
            for &(ref signer_id, did) in &config.signing_keys {
                // Direct storage change for attaching some signing keys to identities
                let signer_key = AccountKey::try_from(signer_id.encode()).unwrap();
                assert!(<DidRecords>::contains_key(did), "Identity does not exist");
                assert!(
                    <Module<T>>::can_key_be_linked_to_did(&signer_key, SignatoryType::External),
                    "Signing key already linked"
                );
                <MultiPurposeNonce>::mutate(|n| *n += 1_u64);
                <Module<T>>::link_key_to_did(&signer_key, SignatoryType::External, did);
                <DidRecords>::mutate(did, |record| {
                    (*record).add_signing_items(&[SigningItem::from(signer_key.clone())]);
                });
                <Module<T>>::deposit_event(RawEvent::SigningItemsAdded(
                    did,
                    [SigningItem::from(signer_key.clone())].to_vec(),
                ));
            }
        });
    }
}

decl_module! {
    /// The module declaration.
    pub struct Module<T: Trait> for enum Call where origin: T::Origin {

        type Error = Error<T>;

        // Initializing events
        // this is needed only if you are using events in your module
        fn deposit_event() = default;

        // TODO: Remove this function before mainnet. cdd_register_did should be used instead.
        /// Register a new did with a CDD claim for the caller.
        pub fn register_did(origin, signing_items: Vec<SigningItem>) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let signer = Signatory::from(AccountKey::try_from(sender.encode())?);
            Self::_register_did(
                sender,
                signing_items,
                Some((&signer, ProtocolOp::IdentityRegisterDid))
            )?;
            Ok(())
        }

        /// Register `target_account` with a new Identity.
        ///
        /// # Failure
        /// - `origin` has to be a active CDD provider. Inactive CDD providers cannot add new
        /// claims.
        /// - `target_account` (master key of the new Identity) can be linked to just one and only
        /// one identity.
        /// - External signing keys can be linked to just one identity.
        ///
        /// # Weight
        /// `400_000 + 60_000 * signing_items.len()`
        #[weight = FunctionOf(
            |(_, _, items): (&T::AccountId, &Option<T::Moment>, &Vec<SigningItem>)| {
                400_000 + 60_000 * u32::try_from(items.len()).unwrap_or_default()
            },
            DispatchClass::Normal,
            true
        )]
        pub fn cdd_register_did(
            origin,
            target_account: T::AccountId,
            cdd_claim_expiry: Option<T::Moment>,
            signing_items: Vec<SigningItem>
        ) -> DispatchResult {
            // Sender has to be part of CDDProviders
            let cdd_sender = ensure_signed(origin)?;
            let cdd_key = AccountKey::try_from(cdd_sender.encode())?;
            let cdd_id = Context::current_identity_or::<Self>(&cdd_key)?;

            let cdd_providers = T::CddServiceProviders::get_members();
            ensure!(cdd_providers.contains(&cdd_id), Error::<T>::UnAuthorizedCddProvider);
            // Register Identity and add claim.
            let new_id = Self::_register_did(
                target_account,
                signing_items,
                Some((&Signatory::AccountKey(cdd_key), ProtocolOp::IdentityCddRegisterDid))
            )?;
            Self::unsafe_add_claim(new_id, Claim::CustomerDueDiligence, cdd_id, cdd_claim_expiry);
            Ok(())
        }

        /// It invalidates any claim generated by `cdd` from `disable_from` timestamps.
        /// You can also define an expiration time, which will invalidate all claims generated by
        /// that `cdd` and remove it as CDD member group.
        #[weight = SimpleDispatchInfo::FixedNormal(400_000)]
        pub fn invalidate_cdd_claims(
            origin,
            cdd: IdentityId,
            disable_from: T::Moment,
            expiry: Option<T::Moment>,
        ) -> DispatchResult {
            ensure_root(origin)?;

            let now = <pallet_timestamp::Module<T>>::get();
            ensure!(
                T::CddServiceProviders::get_valid_members_at(now).contains(&cdd),
                Error::<T>::UnAuthorizedCddProvider);

            T::CddServiceProviders::disable_member(cdd, expiry, Some(disable_from))?;
            Self::deposit_event(RawEvent::CddClaimsInvalidated(cdd, disable_from));
            Ok(())
        }

        /// Removes specified signing keys of a DID if present.
        ///
        /// # Failure
        /// It can only called by master key owner.
        ///
        /// # Weight
        /// `400_000 + 60_000 * signers_to_remove.len()`
        #[weight = FunctionOf(
            |(items,): (&Vec<Signatory>,)| {
                400_000 + 60_000 * u32::try_from(items.len()).unwrap_or_default()
            },
            DispatchClass::Normal,
            true
        )]
        pub fn remove_signing_items(origin, signers_to_remove: Vec<Signatory>) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;
            let did = Context::current_identity_or::<Self>(&sender_key)?;
            let _grants_checked = Self::grant_check_only_master_key(&sender_key, did)?;
            let did_sig = Signatory::from(did);

            // Remove links and get all authorization IDs per signer.
            let signer_and_auth_id_list = signers_to_remove.iter().map(|signer| {
                match signer {
                    Signatory::AccountKey(ref key) => Self::unlink_key_from_did(key, did),
                    _ => {}
                };

                // It returns the list of `auth_id` from `did`.
                let auth_ids = <Authorizations<T>>::iter_prefix(signer)
                    .filter_map( |authorization| {
                        if authorization.authorized_by == did_sig {
                            Some(authorization.auth_id)
                        } else {
                            None
                        }
                    })
                    .collect::<Vec<_>>();

                (signer, auth_ids)
            })
            .collect::<Vec<_>>();

            // Remove authorizations
            signer_and_auth_id_list.into_iter().for_each( |(signer, auth_ids)| {
                auth_ids.into_iter().for_each( |auth_id|
                        Self::unsafe_remove_auth( signer, auth_id, &did_sig, true));
            });

            // Update signing keys at Identity.
            <DidRecords>::mutate(did, |record| {
                (*record).remove_signing_items(&signers_to_remove);
            });

            Self::deposit_event(RawEvent::SigningItemsRemoved(did, signers_to_remove));
            Ok(())
        }

        /// Sets a new master key for a DID.
        ///
        /// # Failure
        /// Only called by master key owner.
        #[weight = SimpleDispatchInfo::FixedNormal(300_000)]
        fn set_master_key(origin, new_key: AccountKey) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let sender_key = AccountKey::try_from( sender.encode())?;
            let did = Context::current_identity_or::<Self>(&sender_key)?;
            let _grants_checked = Self::grant_check_only_master_key(&sender_key, did)?;

            ensure!(
                Self::can_key_be_linked_to_did(&new_key, SignatoryType::External),
                Error::<T>::AlreadyLinked
            );
            T::ProtocolFee::charge_fee(
                &Signatory::AccountKey(sender_key),
                ProtocolOp::IdentitySetMasterKey
            )?;
            <DidRecords>::mutate(did,
            |record| {
                (*record).master_key = new_key.clone();
            });

            Self::deposit_event(RawEvent::MasterKeyUpdated(did, sender_key, new_key));
            Ok(())
        }

        /// Call this with the new master key. By invoking this method, caller accepts authorization
        /// with the new master key. If a CDD service provider approved this change, master key of
        /// the DID is updated.
        ///
        /// # Arguments
        /// * `owner_auth_id` Authorization from the owner who initiated the change
        /// * `cdd_auth_id` Authorization from a CDD service provider
        #[weight = SimpleDispatchInfo::FixedNormal(500_000)]
        pub fn accept_master_key(origin, rotation_auth_id: u64, optional_cdd_auth_id: Option<u64>) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let sender_key = AccountKey::try_from(sender.encode())?;
            Self::accept_master_key_rotation(sender_key, rotation_auth_id, optional_cdd_auth_id)
        }

        /// Set if CDD authorization is required for updating master key of an identity.
        /// Callable via root (governance)
        ///
        /// # Arguments
        /// * `auth_required` CDD Authorization required or not
        #[weight = SimpleDispatchInfo::FixedNormal(200_000)]
        pub fn change_cdd_requirement_for_mk_rotation(
            origin,
            auth_required: bool,
        ) -> DispatchResult {
            ensure_root(origin)?;
            <CddAuthForMasterKeyRotation>::put(auth_required);
            Self::deposit_event(RawEvent::CddRequirementForMasterKeyUpdated(auth_required));
            Ok(())
        }

        /// Join an identity as a signing key.
        #[weight = SimpleDispatchInfo::FixedNormal(300_000)]
        pub fn join_identity_as_key(origin, auth_id: u64) -> DispatchResult {
            let signer = Signatory::from(AccountKey::try_from(ensure_signed(origin)?.encode())?);
            Self::join_identity(signer, auth_id)
        }

        /// Join an identity as a signing identity.
        #[weight = SimpleDispatchInfo::FixedNormal(300_000)]
        pub fn join_identity_as_identity(origin, auth_id: u64) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;
            let sender_did = Context::current_identity_or::<Self>(&sender_key)?;
            Self::join_identity(Signatory::from(sender_did), auth_id)
        }

        /// Leave the signing key's identity.
        #[weight = SimpleDispatchInfo::FixedNormal(300_000)]
        pub fn leave_identity_as_key(origin) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;
            if let Some(did) = Self::get_identity(&sender_key) {
                return Self::leave_identity(Signatory::from(sender_key), did);
            }
            Ok(())
        }

        /// Leave an identity as a signing identity.
        #[weight = SimpleDispatchInfo::FixedNormal(300_000)]
        pub fn leave_identity_as_identity(origin, did: IdentityId) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;
            let sender_did = Context::current_identity_or::<Self>(&sender_key)?;
            Self::leave_identity(Signatory::from(sender_did), did)
        }

        /// Adds a new claim record or edits an existing one. Only called by did_issuer's signing key.
        #[weight = SimpleDispatchInfo::FixedNormal(400_000)]
        pub fn add_claim(
            origin,
            target: IdentityId,
            claim: Claim,
            expiry: Option<T::Moment>,
        ) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;
            let issuer = Context::current_identity_or::<Self>(&sender_key)?;
            ensure!(<DidRecords>::contains_key(target), Error::<T>::DidMustAlreadyExist);

            match claim {
                Claim::CustomerDueDiligence => Self::unsafe_add_cdd_claim(target, claim, issuer, expiry)?,
                _ => {
                    T::ProtocolFee::charge_fee(
                    &Signatory::AccountKey(sender_key),
                    ProtocolOp::IdentityAddClaim
                    )?;
                    Self::unsafe_add_claim(target, claim, issuer, expiry)
                }
            };
            Ok(())
        }

        /// Adds a new batch of claim records or edits an existing one. Only called by
        /// `did_issuer`'s signing key.
        ///
        /// # Weight
        /// `300_000 + 100_000 * claims.len()`
        #[weight = FunctionOf(
            |(claims,): (&Vec<BatchAddClaimItem<T::Moment>>,)| {
                300_000 + 100_000 * u32::try_from(claims.len()).unwrap_or_default()
            },
            DispatchClass::Normal,
            true
        )]
        pub fn add_claims_batch(
            origin,
            claims: Vec<BatchAddClaimItem<T::Moment>>
        ) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;
            let issuer = Context::current_identity_or::<Self>(&sender_key)?;
            // Check input claims.
            ensure!( claims.iter().all(
                |batch_claim_item| <DidRecords>::contains_key(batch_claim_item.target)),
                Error::<T>::DidMustAlreadyExist);

            let cdd_count: usize = claims
                .iter()
                .filter(|batch_claim_item| match batch_claim_item.claim {
                    Claim::CustomerDueDiligence => true,
                    _ => false,
                })
                .count();
            if cdd_count > 0 {
                let cdd_providers = T::CddServiceProviders::get_members();
                ensure!(cdd_providers.contains(&issuer), Error::<T>::UnAuthorizedCddProvider);
            }

            T::ProtocolFee::charge_fee_batch(
                &Signatory::AccountKey(sender_key),
                ProtocolOp::IdentityAddClaim,
                claims.len() - cdd_count
            )?;
            claims
                .into_iter()
                .for_each(|bci| {
                    Self::unsafe_add_claim(bci.target, bci.claim, issuer, bci.expiry)
                });
            Ok(())
        }

        /// Creates a call on behalf of another DID.
        #[weight = FunctionOf(
            |(_, call): (&IdentityId, &Box<T::Proposal>)| {
                500_000 + call.get_dispatch_info().weight
            },
            |(_, call): (&IdentityId, &Box<T::Proposal>)| {
                call.get_dispatch_info().class
            },
            true
        )]
        fn forwarded_call(origin, target_did: IdentityId, proposal: Box<T::Proposal>) -> DispatchResult {
            let sender = ensure_signed(origin)?;

            // 1. Constraints.
            // 1.1. A valid current identity.
            let current_did = Context::current_identity::<Self>()
                .ok_or_else(||Error::<T>::MissingCurrentIdentity)?;

            // 1.2. Check that current_did is a signing key of target_did
            ensure!(
                Self::is_signer_authorized(current_did, &Signatory::Identity(target_did)),
                Error::<T>::CurrentIdentityCannotBeForwarded
            );

            // 1.3. Check that target_did has a CDD.
            ensure!(Self::has_valid_cdd(target_did), Error::<T>::TargetHasNoCdd);

            // 1.4 charge fee
            let _ = T::ChargeTxFeeTarget::charge_fee(
                proposal.encode().len().try_into().unwrap_or_default(),
                proposal.get_dispatch_info())
                    .map_err(|_| Error::<T>::FailedToChargeFee)?;

            // 2. Actions
            T::CddHandler::set_current_identity(&target_did);

            // Also set current_did roles when acting as a signing key for target_did
            // Re-dispatch call - e.g. to asset::doSomething...
            let new_origin = frame_system::RawOrigin::Signed(sender).into();

            proposal.dispatch(new_origin)
        }

        /// Marks the specified claim as revoked.
        #[weight = SimpleDispatchInfo::FixedNormal(250_000)]
        pub fn revoke_claim(origin,
            target: IdentityId,
            claim: Claim,
        ) -> DispatchResult {
            let sender_key = AccountKey::try_from( ensure_signed(origin)?.encode())?;
            let issuer = Context::current_identity_or::<Self>(&sender_key)?;
            let claim_type = claim.claim_type();
            let scope = claim.as_scope().cloned();

            Self::unsafe_revoke_claim(target, claim_type, issuer, scope);
            Ok(())
        }

        /// Revoke multiple claims in a batch.
        ///
        /// # Arguments
        /// * origin - did issuer
        /// * did_and_claim_data - Vector of the identities & the corresponding claim data whom claim needs to be revoked
        ///
        /// # Weight
        /// `100_000 + 150_000 * claims.len()`
        #[weight = FunctionOf(
            |(claims,): (&Vec<BatchRevokeClaimItem>,)| {
                100_000 + 150_000 * u32::try_from(claims.len()).unwrap_or_default()
            },
            DispatchClass::Normal,
            true
        )]
        pub fn revoke_claims_batch(origin,
            claims: Vec<BatchRevokeClaimItem>
        ) -> DispatchResult {
            let sender_key = AccountKey::try_from( ensure_signed(origin)?.encode())?;
            let issuer = Context::current_identity_or::<Self>(&sender_key)?;

            claims.into_iter()
                .for_each( |bci| {
                    let claim_type = bci.claim.claim_type();
                    let scope = bci.claim.as_scope().cloned();
                    Self::unsafe_revoke_claim(bci.target, claim_type, issuer, scope)
                });
            Ok(())
        }

        /// It sets permissions for an specific `target_key` key.
        /// Only the master key of an identity is able to set signing key permissions.
        ///
        /// # Weight
        /// `400_000 + 30_000 * permissions.len()`
        #[weight = FunctionOf(
            |(_, permissions): (&Signatory, &Vec<Permission>)| {
                400_000 + 30_000 * u32::try_from(permissions.len()).unwrap_or_default()
            },
            DispatchClass::Normal,
            true
        )]
        pub fn set_permission_to_signer(origin, signer: Signatory, permissions: Vec<Permission>) -> DispatchResult {
            let sender_key = AccountKey::try_from( ensure_signed(origin)?.encode())?;
            let did = Context::current_identity_or::<Self>(&sender_key)?;
            let record = Self::grant_check_only_master_key( &sender_key, did)?;

            // You are trying to add a permission to did's master key. It is not needed.
            match signer {
                Signatory::AccountKey(ref key) if record.master_key == *key => Ok(()),
                _ if record.signing_items.iter().any(|si| si.signer == signer) => Self::update_signing_item_permissions(did, &signer, permissions),
                _ => Err(Error::<T>::InvalidSender.into()),
            }
        }

        /// It disables all signing keys at `did` identity.
        ///
        /// # Errors
        ///
        #[weight = SimpleDispatchInfo::FixedNormal(200_000)]
        pub fn freeze_signing_keys(origin) -> DispatchResult {
            Self::set_frozen_signing_key_flags(origin, true)
        }

        /// Re-enables all signing keys of the caller's identity.
        #[weight = SimpleDispatchInfo::FixedNormal(200_000)]
        pub fn unfreeze_signing_keys(origin) -> DispatchResult {
            Self::set_frozen_signing_key_flags(origin, false)
        }

        // TODO: Remove before mainnet launch.
        /// Emits an event with caller's identity.
        #[weight = SimpleDispatchInfo::FixedNormal(50_000)]
        pub fn get_my_did(origin) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;
            let did = Context::current_identity_or::<Self>(&sender_key)?;

            Self::deposit_event(RawEvent::DidStatus(did, sender_key));
            Ok(())
        }

        // TODO: Remove before mainnet launch.
        /// Emits an event with caller's identity and CDD status.
        #[weight = SimpleDispatchInfo::FixedNormal(200_000)]
        pub fn get_cdd_of(_origin, of: T::AccountId) -> DispatchResult {
            let key = AccountKey::try_from(of.encode())?;

            let did_opt = Self::get_identity(&key);
            let has_cdd = did_opt.iter()
                .map(|did| Self::has_valid_cdd(*did))
                .next()
                .unwrap_or_default();

            Self::deposit_event(RawEvent::CddStatus(did_opt, key, has_cdd));
            Ok(())
        }

        // Manage generic authorizations
        /// Adds an authorization.
        #[weight = SimpleDispatchInfo::FixedNormal(200_000)]
        pub fn add_authorization(
            origin,
            target: Signatory,
            authorization_data: AuthorizationData,
            expiry: Option<T::Moment>
        ) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;
            let from_did = Context::current_identity_or::<Self>(&sender_key)?;

            Self::add_auth(Signatory::from(from_did), target, authorization_data, expiry);

            Ok(())
        }

        /// Adds an authorization as a key.
        /// To be used by signing keys that don't have an identity.
        #[weight = SimpleDispatchInfo::FixedNormal(150_000)]
        pub fn add_authorization_as_key(
            origin,
            target: Signatory,
            authorization_data: AuthorizationData,
            expiry: Option<T::Moment>
        ) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;

            Self::add_auth(Signatory::from(sender_key), target, authorization_data, expiry);

            Ok(())
        }

        // Manage generic authorizations
        /// Adds an array of authorizations.
        ///
        /// # Weight
        /// `150_000 + 50_000 * auths.len()`
        #[weight = FunctionOf(
            |(auths,): (&Vec<(Signatory, AuthorizationData, Option<T::Moment>)>,)| {
                150_000 + 50_000 * u32::try_from(auths.len()).unwrap_or_default()
            },
            DispatchClass::Normal,
            true
        )]
        pub fn batch_add_authorization(
            origin,
            // Vec<(target_did, auth_data, expiry)>
            auths: Vec<(Signatory, AuthorizationData, Option<T::Moment>)>
        ) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;
            let from_did = Context::current_identity_or::<Self>(&sender_key)?;

            for auth in auths {
                Self::add_auth(Signatory::from(from_did), auth.0, auth.1, auth.2);
            }

            Ok(())
        }

        /// Removes an authorization.
        #[weight = SimpleDispatchInfo::FixedNormal(250_000)]
        pub fn remove_authorization(
            origin,
            target: Signatory,
            auth_id: u64
        ) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;
            let from_did = Context::current_identity_or::<Self>(&sender_key)?;

            ensure!(
                <Authorizations<T>>::contains_key(target, auth_id),
                Error::<T>::AuthorizationDoesNotExist
            );
            let auth = <Authorizations<T>>::get(target, auth_id);
            let revoked = auth.authorized_by.eq_either(&from_did, &sender_key);
            ensure!(
                revoked || target.eq_either(&from_did, &sender_key),
                Error::<T>::Unauthorized
            );
            Self::unsafe_remove_auth(&target, auth_id, &auth.authorized_by, revoked);

            Ok(())
        }

        /// Removes an array of authorizations.
        ///
        /// # Weight
        /// `150_000 + 50_000 * auths.len()`
        #[weight = FunctionOf(
            |(auths,): (&Vec<AuthIdentifier>,)| {
                150_000 + 100_000 * u32::try_from(auths.len()).unwrap_or_default()
            },
            DispatchClass::Normal,
            true
        )]
        pub fn batch_remove_authorization(
            origin,
            auth_identifiers: Vec<AuthIdentifier>
        ) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;
            let from_did = Context::current_identity_or::<Self>(&sender_key)?;
            let mut auths = Vec::with_capacity(auth_identifiers.len());
            let mut revoked = Vec::with_capacity(auth_identifiers.len());
            for i in 0..auth_identifiers.len() {
                let auth_identifier = &auth_identifiers[i];
                ensure!(
                    <Authorizations<T>>::contains_key(&auth_identifier.0, &auth_identifier.1),
                    Error::<T>::AuthorizationDoesNotExist
                );
                auths.push(<Authorizations<T>>::get(&auth_identifier.0, &auth_identifier.1));
                revoked.push(auths[i].authorized_by.eq_either(&from_did, &sender_key));
                ensure!(
                    revoked[i] || auth_identifier.0.eq_either(&from_did, &sender_key),
                    Error::<T>::Unauthorized
                );
            }

            for i in 0..auth_identifiers.len() {
                let auth_identifier = &auth_identifiers[i];
                Self::unsafe_remove_auth(&auth_identifier.0, auth_identifier.1, &auths[i].authorized_by, revoked[i]);

            }

            Ok(())
        }

        /// Accepts an authorization.
        #[weight = SimpleDispatchInfo::FixedNormal(600_000)]
        pub fn accept_authorization(
            origin,
            auth_id: u64
        ) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;
            let signer_by_key = Signatory::from(sender_key.clone());
            let signer_by_id = Context::current_identity_or::<Self>(&sender_key)
                .map_or_else(
                    |_error| Signatory::from(sender_key),
                    Signatory::from);

            // Get auth by key or by id.
            let auth_signer_opt = if <Authorizations<T>>::contains_key(&signer_by_id, auth_id) {
                let auth = <Authorizations<T>>::get(&signer_by_id, auth_id);
                Some((auth, signer_by_id))
            } else if <Authorizations<T>>::contains_key(signer_by_key, auth_id) {
                let auth = <Authorizations<T>>::get(signer_by_key, auth_id);
                Some((auth, signer_by_key))
            } else {
                None
            };

            let (auth, signer) = auth_signer_opt
                    .ok_or_else(||Error::<T>::AuthorizationDoesNotExist)?;

            match signer {
                Signatory::Identity(did) => {
                    match auth.authorization_data {
                        AuthorizationData::TransferTicker(_) =>
                            T::AcceptTransferTarget::accept_ticker_transfer(did, auth_id),
                        AuthorizationData::TransferAssetOwnership(_) =>
                            T::AcceptTransferTarget::accept_asset_ownership_transfer(did, auth_id),
                        AuthorizationData::AddMultiSigSigner =>
                            T::AddSignerMultiSigTarget::accept_multisig_signer(Signatory::from(did), auth_id),
                        AuthorizationData::JoinIdentity(_) =>
                            Self::join_identity(Signatory::from(did), auth_id),
                        _ => Err(Error::<T>::UnknownAuthorization.into())
                    }
                },
                Signatory::AccountKey(key) => {
                    match auth.authorization_data {
                        AuthorizationData::AddMultiSigSigner =>
                            T::AddSignerMultiSigTarget::accept_multisig_signer(Signatory::from(key), auth_id),
                        AuthorizationData::RotateMasterKey(_identityid) =>
                            Self::accept_master_key_rotation(key , auth_id, None),
                        AuthorizationData::JoinIdentity(_) =>
                            Self::join_identity(Signatory::from(key), auth_id),
                        _ => Err(Error::<T>::UnknownAuthorization.into())
                    }
                }
            }
        }

        /// Accepts an array of authorizations.
        ///
        /// # Weight
        /// `100_000 + 500_000 * auth_ids.len()`
        #[weight = FunctionOf(
            |(auth_ids,): (&Vec<u64>,)| {
                100_000 + 500_000 * u32::try_from(auth_ids.len()).unwrap_or_default()
            },
            DispatchClass::Normal,
            true
        )]
        pub fn batch_accept_authorization(
            origin,
            auth_ids: Vec<u64>
        ) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;
            let signer = Context::current_identity_or::<Self>(&sender_key)
                .map_or_else(
                    |_error| Signatory::from(sender_key),
                    Signatory::from);

            match signer {
                Signatory::Identity(did) => {
                    for auth_id in auth_ids {
                        // NB: Even if an auth is invalid (due to any reason), this batch function does NOT return an error.
                        // It will just skip that particular authorization.

                        if <Authorizations<T>>::contains_key(signer, auth_id) {
                            let auth = <Authorizations<T>>::get(signer, auth_id);

                            // NB: Result is not handled, invalid auths are just ignored to let the batch function continue.
                            let _result = match auth.authorization_data {
                                AuthorizationData::TransferTicker(_) =>
                                    T::AcceptTransferTarget::accept_ticker_transfer(did, auth_id),
                                AuthorizationData::TransferAssetOwnership(_) =>
                                    T::AcceptTransferTarget::accept_asset_ownership_transfer(did, auth_id),
                                AuthorizationData::AddMultiSigSigner =>
                                    T::AddSignerMultiSigTarget::accept_multisig_signer(Signatory::from(did), auth_id),
                                AuthorizationData::JoinIdentity(_) =>
                                    Self::join_identity(Signatory::from(did), auth_id),
                                _ => Err(Error::<T>::UnknownAuthorization.into())
                            };
                        }
                    }
                },
                Signatory::AccountKey(key) => {
                    for auth_id in auth_ids {
                        // NB: Even if an auth is invalid (due to any reason), this batch function does NOT return an error.
                        // It will just skip that particular authorization.

                        if <Authorizations<T>>::contains_key(signer, auth_id) {
                            let auth = <Authorizations<T>>::get(signer, auth_id);

                            //NB: Result is not handled, invalid auths are just ignored to let the batch function continue.
                            let _result = match auth.authorization_data {
                                AuthorizationData::AddMultiSigSigner =>
                                    T::AddSignerMultiSigTarget::accept_multisig_signer(Signatory::from(key), auth_id),
                                AuthorizationData::RotateMasterKey(_identityid) =>
                                    Self::accept_master_key_rotation(key , auth_id, None),
                                AuthorizationData::JoinIdentity(_) =>
                                    Self::join_identity(Signatory::from(key), auth_id),
                                _ => Err(Error::<T>::UnknownAuthorization.into())
                            };
                        }
                    }
                }
            }

            Ok(())
        }

        /// It adds signing keys to target identity `id`.
        /// Keys are directly added to identity because each of them has an authorization.
        ///
        /// Arguments:
        ///     - `origin` Master key of `id` identity.
        ///     - `id` Identity where new signing keys will be added.
        ///     - `additional_keys` New signing items (and their authorization data) to add to target
        ///     identity.
        ///
        /// Failure
        ///     - It can only called by master key owner.
        ///     - Keys should be able to linked to any identity.
        ///
        /// # Weight
        /// `400_000 + 200_000 * auths.len()`
        #[weight = FunctionOf(
            |(_, additional_keys): (&T::Moment, &Vec<SigningItemWithAuth>)| {
                400_000 + 200_000 * u32::try_from(additional_keys.len()).unwrap_or_default()
            },
            DispatchClass::Normal,
            true
        )]
        pub fn add_signing_items_with_authorization(
            origin,
            expires_at: T::Moment,
            additional_keys: Vec<SigningItemWithAuth>
        ) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let sender_key = AccountKey::try_from(sender.encode())?;
            let id = Context::current_identity_or::<Self>(&sender_key)?;
            let _grants_checked = Self::grant_check_only_master_key(&sender_key, id)?;

            // 0. Check expiration
            let now = <pallet_timestamp::Module<T>>::get();
            ensure!(now < expires_at, Error::<T>::AuthorizationExpired);
            let authorization = TargetIdAuthorization {
                target_id: id,
                nonce: Self::offchain_authorization_nonce(id),
                expires_at
            };
            let auth_encoded = authorization.encode();

            // 1. Verify signatures.
            for si_with_auth in additional_keys.iter() {
                let si = &si_with_auth.signing_item;

                // Get account_id from signer
                let account_id_found = match si.signer {
                    Signatory::AccountKey(ref key) =>  Public::try_from(key.as_slice()).ok(),
                    Signatory::Identity(ref id) if <DidRecords>::contains_key(id) => {
                        let master_key = <DidRecords>::get(id).master_key;
                        Public::try_from( master_key.as_slice()).ok()
                    },
                    _ => None
                };

                if let Some(account_id) = account_id_found {
                    if let Signatory::AccountKey(ref key) = si.signer {
                        // 1.1. Constraint 1-to-1 account to DID
                        ensure!(
                            Self::can_key_be_linked_to_did(key, si.signer_type),
                            Error::<T>::AlreadyLinked
                        );
                    }
                    // 1.2. Offchain authorization is not revoked explicitly.
                    let si_signer_authorization = &(si.signer, authorization.clone());
                    ensure!(
                        !Self::is_offchain_authorization_revoked(si_signer_authorization),
                        Error::<T>::AuthorizationHasBeenRevoked
                    );
                    // 1.3. Verify the signature.
                    let signature = AnySignature::from(Signature::from_h512(si_with_auth.auth_signature));
                    ensure!(
                        signature.verify(auth_encoded.as_slice(), &account_id),
                        Error::<T>::InvalidAuthorizationSignature
                    );
                } else {
                    return Err(Error::<T>::InvalidAccountKey.into());
                }
            }
            // 1.999. Charge the fee.
            T::ProtocolFee::charge_fee_batch(
                &Signatory::AccountKey(sender_key),
                ProtocolOp::IdentityAddSigningItemsWithAuthorization,
                additional_keys.len()
            )?;
            // 2.1. Link keys to identity
            let additional_keys_si = additional_keys.into_iter()
                .map( |si_with_auth| si_with_auth.signing_item)
                .collect::<Vec<_>>();

            additional_keys_si.iter().for_each( |si| {
                if let Signatory::AccountKey(ref key) = si.signer {
                    Self::link_key_to_did( key, si.signer_type, id);
                }
            });
            // 2.2. Update that identity information and its offchain authorization nonce.
            <DidRecords>::mutate(id, |record| {
                (*record).add_signing_items(&additional_keys_si[..]);
            });
            <OffChainAuthorizationNonce>::mutate(id, |offchain_nonce| {
                *offchain_nonce = authorization.nonce + 1;
            });

            Self::deposit_event(RawEvent::SigningItemsAdded(id, additional_keys_si));

            Ok(())
        }

        /// It revokes the `auth` off-chain authorization of `signer`. It only takes effect if
        /// the authorized transaction is not yet executed.
        #[weight = SimpleDispatchInfo::FixedNormal(100_000)]
        pub fn revoke_offchain_authorization(origin, signer: Signatory, auth: TargetIdAuthorization<T::Moment>) -> DispatchResult {
            let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;

            match signer {
                Signatory::AccountKey(ref key) => {
                    ensure!(sender_key == *key, Error::<T>::KeyNotAllowed);
                }
                Signatory::Identity(id) => {
                    ensure!(Self::is_master_key(id, &sender_key), Error::<T>::NotMasterKey);
                }
            }

            Self::deposit_event(RawEvent::OffChainAuthorizationRevoked(auth.target_id, signer));
            <RevokeOffChainAuthorization<T>>::insert((signer,auth), true);
            Ok(())
        }
    }
}

decl_error! {
    pub enum Error for Module<T: Trait> {
        /// One signing key can only belong to one DID
        AlreadyLinked,
        /// Missing current identity on the transaction
        MissingCurrentIdentity,
        /// Sender is not part of did's signing keys
        InvalidSender,
        /// No did linked to the user
        NoDIDFound,
        /// Signatory is not pre authorized by the identity
        Unauthorized,
        /// Given authorization is not pre-known
        UnknownAuthorization,
        /// Account Id cannot be extracted from signer
        InvalidAccountKey,
        /// Only CDD service providers are allowed.
        UnAuthorizedCddProvider,
        /// An invalid authorization from the owner.
        InvalidAuthorizationFromOwner,
        /// An invalid authorization from the CDD provider.
        InvalidAuthorizationFromCddProvider,
        /// The authorization to change the key was not from the owner of the master key.
        KeyChangeUnauthorized,
        /// Attestation was not by a CDD service provider.
        NotCddProviderAttestation,
        /// Authorizations are not for the same DID.
        AuthorizationsNotForSameDids,
        /// The DID must already exist.
        DidMustAlreadyExist,
        /// The Claim issuer DID must already exist.
        ClaimIssuerDidMustAlreadyExist,
        /// Sender must hold a claim issuer's signing key.
        SenderMustHoldClaimIssuerKey,
        /// Current identity cannot be forwarded, it is not a signing key of target identity.
        CurrentIdentityCannotBeForwarded,
        /// The authorization does not exist.
        AuthorizationDoesNotExist,
        /// The offchain authorization has expired.
        AuthorizationExpired,
        /// The master key is already linked to an identity.
        MasterKeyAlreadyLinked,
        /// The target DID has no valid CDD.
        TargetHasNoCdd,
        /// Authorization has been explicitly revoked.
        AuthorizationHasBeenRevoked,
        /// An invalid authorization signature.
        InvalidAuthorizationSignature,
        /// This key is not allowed to execute a given operation.
        KeyNotAllowed,
        /// Only the master key is allowed to revoke an Identity Signatory off-chain authorization.
        NotMasterKey,
        /// The DID does not exist.
        DidDoesNotExist,
        /// The DID already exists.
        DidAlreadyExists,
        /// The signing keys contain the master key.
        SigningKeysContainMasterKey,
        /// Couldn't charge fee for the transaction
        FailedToChargeFee,
        /// Signer is not a signing key of the provided identity
        NotASigner,
    }
}

impl<T: Trait> Module<T> {
    /// Accepts an auth to join an identity as a signer
    pub fn join_identity(signer: Signatory, auth_id: u64) -> DispatchResult {
        ensure!(
            <Authorizations<T>>::contains_key(signer, auth_id),
            AuthorizationError::Invalid
        );

        let auth = <Authorizations<T>>::get(signer, auth_id);

        let identity_to_join = match auth.authorization_data {
            AuthorizationData::JoinIdentity(identity) => Ok(identity),
            _ => Err(AuthorizationError::Invalid),
        }?;

        ensure!(
            <DidRecords>::contains_key(&identity_to_join),
            "Identity does not exist"
        );

        let master = Self::did_records(&identity_to_join).master_key;

        Self::consume_auth(Signatory::from(master), signer, auth_id)?;

        Self::unsafe_join_identity(identity_to_join, signer)
    }

    /// Joins an identity as signer
    pub fn unsafe_join_identity(identity_to_join: IdentityId, signer: Signatory) -> DispatchResult {
        if let Signatory::AccountKey(key) = signer {
            if !Self::can_key_be_linked_to_did(&key, SignatoryType::External) {
                ensure!(
                    Self::get_identity(&key) == Some(identity_to_join),
                    Error::<T>::AlreadyLinked
                )
            } else {
                T::ProtocolFee::charge_fee(
                    &Signatory::Identity(identity_to_join),
                    ProtocolOp::IdentityAddSigningItemsWithAuthorization,
                )?;
                Self::link_key_to_did(&key, SignatoryType::External, identity_to_join);
            }
        } else {
            T::ProtocolFee::charge_fee(
                &Signatory::Identity(identity_to_join),
                ProtocolOp::IdentityAddSigningItemsWithAuthorization,
            )?;
        }
        <DidRecords>::mutate(identity_to_join, |identity| {
            identity.add_signing_items(&[SigningItem::new(signer, vec![])]);
        });

        Self::deposit_event(RawEvent::SigningItemsAdded(
            identity_to_join,
            [SigningItem::new(signer, vec![])].to_vec(),
        ));

        Ok(())
    }

    /// Adds an authorization.
    pub fn add_auth(
        from: Signatory,
        target: Signatory,
        authorization_data: AuthorizationData,
        expiry: Option<T::Moment>,
    ) -> u64 {
        let new_nonce = Self::multi_purpose_nonce() + 1u64;
        <MultiPurposeNonce>::put(&new_nonce);

        let auth = Authorization {
            authorization_data: authorization_data.clone(),
            authorized_by: from,
            expiry,
            auth_id: new_nonce,
        };

        <Authorizations<T>>::insert(target, new_nonce, auth);
        <AuthorizationsGiven>::insert(from, new_nonce, target);

        // This event is split in order to help the event harvesters.
        match from {
            Signatory::Identity(id) => Self::deposit_event(RawEvent::AuthorizationAddedByIdentity(
                id,
                target.as_identity().cloned(),
                target.as_account_key().cloned(),
                new_nonce,
                authorization_data,
                expiry,
            )),
            Signatory::AccountKey(key) => Self::deposit_event(RawEvent::AuthorizationAddedByKey(
                key,
                target.as_identity().cloned(),
                target.as_account_key().cloned(),
                new_nonce,
                authorization_data,
                expiry,
            )),
        };

        new_nonce
    }

    /// Removes any authorization. No questions asked.
    /// NB: Please do all the required checks before calling this function.
    pub fn unsafe_remove_auth(
        target: &Signatory,
        auth_id: u64,
        authorizer: &Signatory,
        revoked: bool,
    ) {
        <Authorizations<T>>::remove(target, auth_id);
        <AuthorizationsGiven>::remove(authorizer, auth_id);
        if revoked {
            Self::deposit_event(RawEvent::AuthorizationRevoked(
                target.as_identity().cloned(),
                target.as_account_key().cloned(),
                auth_id,
            ));
        } else {
            Self::deposit_event(RawEvent::AuthorizationRejected(
                target.as_identity().cloned(),
                target.as_account_key().cloned(),
                auth_id,
            ));
        }
    }

    /// Consumes an authorization.
    /// Checks if the auth has not expired and the caller is authorized to consume this auth.
    pub fn consume_auth(from: Signatory, target: Signatory, auth_id: u64) -> DispatchResult {
        ensure!(
            <Authorizations<T>>::contains_key(target, auth_id),
            AuthorizationError::Invalid
        );
        let auth = <Authorizations<T>>::get(target, auth_id);
        ensure!(auth.authorized_by == from, AuthorizationError::Unauthorized);
        if let Some(expiry) = auth.expiry {
            let now = <pallet_timestamp::Module<T>>::get();
            ensure!(expiry > now, AuthorizationError::Expired);
        }

        <Authorizations<T>>::remove(target, auth_id);
        <AuthorizationsGiven>::remove(auth.authorized_by, auth_id);

        Self::deposit_event(RawEvent::AuthorizationConsumed(
            target.as_identity().cloned(),
            target.as_account_key().cloned(),
            auth_id,
        ));
        Ok(())
    }

    /// Fetches a particular authorization.
    pub fn get_authorization(target: Signatory, auth_id: u64) -> Authorization<T::Moment> {
        <Authorizations<T>>::get(target, auth_id)
    }

    /// Fetches a particular link.
    pub fn get_link(target: Signatory, link_id: u64) -> Link<T::Moment> {
        <Links<T>>::get(target, link_id)
    }

    /// Adds a link to a key or an identity.
    /// NB: Please do all the required checks before calling this function.
    pub fn add_link(target: Signatory, link_data: LinkData, expiry: Option<T::Moment>) -> u64 {
        let new_nonce = Self::multi_purpose_nonce() + 1u64;
        <MultiPurposeNonce>::put(&new_nonce);

        let link = Link {
            link_data: link_data.clone(),
            expiry,
            link_id: new_nonce,
        };

        <Links<T>>::insert(target, new_nonce, link);

        Self::deposit_event(RawEvent::LinkAdded(
            target.as_identity().cloned(),
            target.as_account_key().cloned(),
            new_nonce,
            link_data,
            expiry,
        ));
        new_nonce
    }

    /// Remove a link (if it exists) from a key or identity.
    /// NB: Please do all the required checks before calling this function.
    pub fn remove_link(target: Signatory, link_id: u64) {
        if <Links<T>>::contains_key(target, link_id) {
            <Links<T>>::remove(target, link_id);
            Self::deposit_event(RawEvent::LinkRemoved(
                target.as_identity().cloned(),
                target.as_account_key().cloned(),
                link_id,
            ));
        }
    }

    /// Update link data (if it exists) from a key or identity.
    /// NB: Please do all the required checks before calling this function.
    pub fn update_link(target: Signatory, link_id: u64, link_data: LinkData) {
        if <Links<T>>::contains_key(target, link_id) {
            <Links<T>>::mutate(target, link_id, |link| link.link_data = link_data);
            Self::deposit_event(RawEvent::LinkUpdated(
                target.as_identity().cloned(),
                target.as_account_key().cloned(),
                link_id,
            ));
        }
    }

    /// Accepts a master key rotation.
    fn accept_master_key_rotation(
        sender_key: AccountKey,
        rotation_auth_id: u64,
        optional_cdd_auth_id: Option<u64>,
    ) -> DispatchResult {
        let signer = Signatory::from(sender_key);
        // ensure authorization is present
        ensure!(
            <Authorizations<T>>::contains_key(signer, rotation_auth_id),
            Error::<T>::InvalidAuthorizationFromOwner
        );

        // Accept authorization from the owner
        let rotation_auth = <Authorizations<T>>::get(signer, rotation_auth_id);

        if let AuthorizationData::RotateMasterKey(rotation_for_did) =
            rotation_auth.authorization_data
        {
            // Ensure the request was made by the owner of master key
            match rotation_auth.authorized_by {
                Signatory::AccountKey(key) => {
                    let master_key = <DidRecords>::get(rotation_for_did).master_key;
                    ensure!(key == master_key, Error::<T>::KeyChangeUnauthorized);
                }
                _ => return Err(Error::<T>::UnknownAuthorization.into()),
            };
            // consume owner's authorization
            Self::consume_auth(rotation_auth.authorized_by, signer, rotation_auth_id)?;
            Self::unsafe_master_key_rotation(sender_key, rotation_for_did, optional_cdd_auth_id)
        } else {
            Err(Error::<T>::UnknownAuthorization.into())
        }
    }

    /// Processes master key rotation.
    pub fn unsafe_master_key_rotation(
        sender_key: AccountKey,
        rotation_for_did: IdentityId,
        optional_cdd_auth_id: Option<u64>,
    ) -> DispatchResult {
        // Aceept authorization from CDD service provider
        if Self::cdd_auth_for_master_key_rotation() {
            let cdd_auth_id = optional_cdd_auth_id
                .ok_or_else(|| Error::<T>::InvalidAuthorizationFromCddProvider)?;

            let signer = Signatory::from(sender_key);
            ensure!(
                <Authorizations<T>>::contains_key(signer, cdd_auth_id),
                Error::<T>::InvalidAuthorizationFromCddProvider
            );
            let cdd_auth = <Authorizations<T>>::get(signer, cdd_auth_id);

            match cdd_auth.authorization_data {
                AuthorizationData::AttestMasterKeyRotation(ref attestation_for_did) => {
                    // Attestor must be a CDD service provider
                    let cdd_provider_did = match cdd_auth.authorized_by {
                        Signatory::AccountKey(ref key) => Self::get_identity(key),
                        Signatory::Identity(id) => Some(id),
                    };

                    let id = cdd_provider_did.ok_or_else(|| Error::<T>::NoDIDFound)?;
                    ensure!(
                        T::CddServiceProviders::is_member(&id),
                        Error::<T>::NotCddProviderAttestation
                    );

                    // Make sure authorizations are for the same DID
                    ensure!(
                        rotation_for_did == *attestation_for_did,
                        Error::<T>::AuthorizationsNotForSameDids
                    );

                    // consume CDD service provider's authorization
                    Self::consume_auth(cdd_auth.authorized_by, signer, cdd_auth_id)?;
                }
                _ => return Err(Error::<T>::UnknownAuthorization.into()),
            }
        }

        // Replace master key of the owner that initiated key rotation
        let old_master_key = Self::did_records(&rotation_for_did).master_key;
        <DidRecords>::mutate(&rotation_for_did, |record| {
            Self::unlink_key_from_did(&(*record).master_key, rotation_for_did);
            (*record).master_key = sender_key;
        });

        Self::deposit_event(RawEvent::MasterKeyUpdated(
            rotation_for_did,
            old_master_key,
            sender_key,
        ));
        Ok(())
    }

    /// Updates permissions of signing items.
    /// Private and not sanitized function. It is designed to be used internally by
    /// others sanitezed functions.
    fn update_signing_item_permissions(
        target_did: IdentityId,
        signer: &Signatory,
        mut permissions: Vec<Permission>,
    ) -> DispatchResult {
        // Remove duplicates.
        permissions.sort();
        permissions.dedup();

        let mut new_s_item: Option<SigningItem> = None;

        <DidRecords>::mutate(target_did, |record| {
            if let Some(mut signing_item) = (*record)
                .signing_items
                .iter()
                .find(|si| si.signer == *signer)
                .cloned()
            {
                swap(&mut signing_item.permissions, &mut permissions);
                (*record).signing_items.retain(|si| si.signer != *signer);
                (*record).signing_items.push(signing_item.clone());
                new_s_item = Some(signing_item);
            }
        });

        if let Some(s) = new_s_item {
            Self::deposit_event(RawEvent::SigningPermissionsUpdated(
                target_did,
                s,
                permissions,
            ));
        }
        Ok(())
    }

    /// It checks if `key` is a signing key of `did` identity.
    /// # IMPORTANT
    /// If signing keys are frozen this function always returns false.
    /// Master key cannot be frozen.
    pub fn is_signer_authorized(did: IdentityId, signer: &Signatory) -> bool {
        let record = <DidRecords>::get(did);

        // Check master id or key
        match signer {
            Signatory::AccountKey(ref signer_key) if record.master_key == *signer_key => true,
            Signatory::Identity(ref signer_id) if did == *signer_id => true,
            _ => {
                // Check signing items if DID is not frozen.
                !Self::is_did_frozen(did)
                    && record.signing_items.iter().any(|si| si.signer == *signer)
            }
        }
    }

    /// It checks if `key` is a signing key of `did` identity.
    pub fn is_signer(did: IdentityId, signer: &Signatory) -> bool {
        let record = <DidRecords>::get(did);
        record.signing_items.iter().any(|si| si.signer == *signer)
    }

    /// Checks if signer has correct permissions.
    fn is_signer_authorized_with_permissions(
        did: IdentityId,
        signer: &Signatory,
        permissions: Vec<Permission>,
    ) -> bool {
        let record = <DidRecords>::get(did);

        match signer {
            Signatory::AccountKey(ref signer_key) if record.master_key == *signer_key => true,
            Signatory::Identity(ref signer_id) if did == *signer_id => true,
            _ => {
                if !Self::is_did_frozen(did) {
                    if let Some(signing_item) =
                        record.signing_items.iter().find(|&si| &si.signer == signer)
                    {
                        // It retruns true if all requested permission are in this signing item.
                        return permissions.iter().all(|required_permission| {
                            signing_item.has_permission(*required_permission)
                        });
                    }
                }
                // Signatory is not part of signing items of `did`, or
                // Did is frozen.
                false
            }
        }
    }

    /// Use `did` as reference.
    pub fn is_master_key(did: IdentityId, key: &AccountKey) -> bool {
        key == &<DidRecords>::get(did).master_key
    }

    /// It returns true if `id_claim` is not expired at `moment`.
    #[inline]
    fn is_identity_claim_not_expired_at(id_claim: &IdentityClaim, moment: T::Moment) -> bool {
        if let Some(expiry) = id_claim.expiry {
            expiry > moment.saturated_into::<u64>()
        } else {
            true
        }
    }

    /// It fetches an specific `claim_type` claim type for target identity `id`, which was issued
    /// by `issuer`.
    /// It only returns non-expired claims.
    pub fn fetch_claim(
        id: IdentityId,
        claim_type: ClaimType,
        issuer: IdentityId,
        scope: Option<Scope>,
    ) -> Option<IdentityClaim> {
        let now = <pallet_timestamp::Module<T>>::get();

        Self::fetch_base_claim_with_issuer(id, claim_type, issuer, scope)
            .into_iter()
            .filter(|c| Self::is_identity_claim_not_expired_at(c, now))
            .next()
    }

    /// See `Self::fetch_cdd`.
    #[inline]
    pub fn has_valid_cdd(claim_for: IdentityId) -> bool {
        let trusted_cdd_providers = T::CddServiceProviders::get_members();
        // It will never happen in production but helpful during testing.
        // TODO: Remove this condition
        if trusted_cdd_providers.len() == 0 {
            return true;
        }

        Self::fetch_cdd(claim_for, T::Moment::zero()).is_some()
    }

    /// It returns the CDD identity which issued the current valid CDD claim for `claim_for`
    /// identity.
    /// # Parameters
    /// * `leeway` : This leeway is added to now() before check if claim is expired.
    ///
    /// # Safety
    ///
    /// No state change is allowed in this function because this function is used within the RPC
    /// calls.
    pub fn fetch_cdd(claim_for: IdentityId, leeway: T::Moment) -> Option<IdentityId> {
        let exp_with_leeway = <pallet_timestamp::Module<T>>::get()
            .checked_add(&leeway)
            .unwrap_or_default();

        let active_cdds = T::CddServiceProviders::get_active_members();
        let inactive_not_expired_cdds = T::CddServiceProviders::get_inactive_members()
            .into_iter()
            .filter(|cdd| !T::CddServiceProviders::is_member_expired(cdd, exp_with_leeway))
            .collect::<Vec<_>>();

        Self::fetch_base_claims(claim_for, ClaimType::CustomerDueDiligence)
            .filter(|id_claim| {
                Self::is_identity_cdd_claim_valid(
                    id_claim,
                    exp_with_leeway,
                    &active_cdds,
                    &inactive_not_expired_cdds,
                )
            })
            .map(|id_claim| id_claim.claim_issuer)
            .next()
    }

    /// A CDD claims is considered valid if:
    /// * Claim is not expired at `exp_with_leeway` moment.
    /// * Its issuer is valid, that means:
    ///   * Issuer is an active CDD provider, or
    ///   * Issuer is an inactive CDD provider but claim was updated/created before that it was
    ///   deactivated.
    fn is_identity_cdd_claim_valid(
        id_claim: &IdentityClaim,
        exp_with_leeway: T::Moment,
        active_cdds: &[IdentityId],
        inactive_not_expired_cdds: &[InactiveMember<T::Moment>],
    ) -> bool {
        let systematic_cdds = [
            SystematicIssuers::TreasuryModule.as_id(),
            SystematicIssuers::Committee.as_id(),
            SystematicIssuers::CDDProvider.as_id(),
        ];
        Self::is_identity_claim_not_expired_at(id_claim, exp_with_leeway)
            && (active_cdds.contains(&id_claim.claim_issuer)
                || systematic_cdds.contains(&id_claim.claim_issuer)
                || inactive_not_expired_cdds
                    .iter()
                    .filter(|cdd| cdd.id == id_claim.claim_issuer)
                    .any(|cdd| {
                        id_claim.last_update_date < cdd.deactivated_at.saturated_into::<u64>()
                    }))
    }

    /// It iterates over all claims of type `claim_type` for target `id` identity.
    /// Please note that it could return expired claims.
    fn fetch_base_claims<'a>(
        target: IdentityId,
        claim_type: ClaimType,
    ) -> impl Iterator<Item = IdentityClaim> + 'a {
        let pk = Claim1stKey { target, claim_type };
        <Claims>::iter_prefix(pk)
    }

    /// It fetches an specific `claim_type` claim type for target identity `id`, which was issued
    /// by `issuer`.
    fn fetch_base_claim_with_issuer(
        target: IdentityId,
        claim_type: ClaimType,
        issuer: IdentityId,
        scope: Option<Scope>,
    ) -> Option<IdentityClaim> {
        let pk = Claim1stKey { target, claim_type };
        let sk = Claim2ndKey { issuer, scope };

        if <Claims>::contains_key(&pk, &sk) {
            Some(<Claims>::get(&pk, &sk))
        } else {
            None
        }
    }

    /// It checks that `sender_key` is the master key of `did` Identifier and that
    /// did exists.
    /// # Return
    /// A result object containing the `DidRecord` of `did`.
    pub fn grant_check_only_master_key(
        sender_key: &AccountKey,
        did: IdentityId,
    ) -> sp_std::result::Result<DidRecord, Error<T>> {
        ensure!(<DidRecords>::contains_key(did), Error::<T>::DidDoesNotExist);
        let record = <DidRecords>::get(did);
        ensure!(*sender_key == record.master_key, Error::<T>::KeyNotAllowed);
        Ok(record)
    }

    /// It checks if `key` is the master key or signing key of any IdentityId.
    /// Please note that _frozen signing keys_ are not lined to the frozen identity temporary.
    ///
    /// # Return
    ///
    /// An Option object containing the `IdentityId` that belongs to the key.
    pub fn get_identity(key: &AccountKey) -> Option<IdentityId> {
        if let Some(linked_key_info) = <KeyToIdentityIds>::get(key) {
            let id = match linked_key_info {
                LinkedKeyInfo::Unique(id)
                    if !Self::is_did_frozen(id) || Self::is_master_key(id, key) =>
                {
                    Some(id)
                }
                _ => None,
            };

            return id;
        }
        None
    }

    /// It freezes/unfreezes the target `did` identity.
    ///
    /// # Errors
    /// Only master key can freeze/unfreeze an identity.
    fn set_frozen_signing_key_flags(origin: T::Origin, freeze: bool) -> DispatchResult {
        let sender_key = AccountKey::try_from(ensure_signed(origin)?.encode())?;
        let did = Context::current_identity_or::<Self>(&sender_key)?;
        let _grants_checked = Self::grant_check_only_master_key(&sender_key, did)?;

        if freeze {
            <IsDidFrozen>::insert(&did, true);
            Self::deposit_event(RawEvent::SigningKeysFrozen(did));
        } else {
            <IsDidFrozen>::remove(&did);
            Self::deposit_event(RawEvent::SigningKeysUnfrozen(did));
        }
        Ok(())
    }

    /// It checks that any sternal account can only be associated with at most one.
    /// Master keys are considered as external accounts.
    pub fn can_key_be_linked_to_did(key: &AccountKey, signer_type: SignatoryType) -> bool {
        if let Some(linked_key_info) = <KeyToIdentityIds>::get(key) {
            match linked_key_info {
                LinkedKeyInfo::Unique(..) => false,
                LinkedKeyInfo::Group(..) => signer_type != SignatoryType::External,
            }
        } else {
            true
        }
    }

    /// It links `key` key to `did` identity as a `key_type` type.
    /// # Errors
    /// This function can be used if `can_key_be_linked_to_did` returns true. Otherwise, it will do
    /// nothing.
    fn link_key_to_did(key: &AccountKey, key_type: SignatoryType, did: IdentityId) {
        if let Some(linked_key_info) = <KeyToIdentityIds>::get(key) {
            if let LinkedKeyInfo::Group(mut dids) = linked_key_info {
                if !dids.contains(&did) && key_type != SignatoryType::External {
                    dids.push(did);
                    dids.sort();

                    <KeyToIdentityIds>::insert(key, LinkedKeyInfo::Group(dids));
                }
            }
        } else {
            // AccountKey is not yet linked to any identity, so no constraints.
            let linked_key_info = match key_type {
                SignatoryType::External => LinkedKeyInfo::Unique(did),
                _ => LinkedKeyInfo::Group(vec![did]),
            };
            <KeyToIdentityIds>::insert(key, linked_key_info);
        }
    }

    /// It unlinks the `key` key from `did`.
    /// If there is no more associated identities, its full entry is removed.
    fn unlink_key_from_did(key: &AccountKey, did: IdentityId) {
        if let Some(linked_key_info) = <KeyToIdentityIds>::get(key) {
            match linked_key_info {
                LinkedKeyInfo::Unique(did_linked) => {
                    if did_linked == did {
                        <KeyToIdentityIds>::remove(key)
                    }
                }
                LinkedKeyInfo::Group(mut dids) => {
                    dids.retain(|ref_did| *ref_did != did);
                    if dids.is_empty() {
                        <KeyToIdentityIds>::remove(key);
                    } else {
                        <KeyToIdentityIds>::insert(key, LinkedKeyInfo::Group(dids));
                    }
                }
            }
        }
    }

    /// It registers a did for a new asset. Only called by create_asset function.
    pub fn register_asset_did(ticker: &Ticker) -> DispatchResult {
        let did = Self::get_token_did(ticker)?;
        // Making sure there's no pre-existing entry for the DID
        // This should never happen but just being defensive here
        ensure!(
            !<DidRecords>::contains_key(did),
            Error::<T>::DidAlreadyExists
        );
        <DidRecords>::insert(did, DidRecord::default());
        Self::deposit_event(RawEvent::AssetDidRegistered(did, *ticker));
        Ok(())
    }

    /// IMPORTANT: No state change is allowed in this function
    /// because this function is used within the RPC calls
    /// It is a helper function that can be used to get did for any asset
    pub fn get_token_did(ticker: &Ticker) -> StdResult<IdentityId, &'static str> {
        let mut buf = Vec::new();
        buf.extend_from_slice(&SECURITY_TOKEN.encode());
        buf.extend_from_slice(&ticker.encode());
        IdentityId::try_from(T::Hashing::hash(&buf[..]).as_ref())
    }

    /// Registers a did without adding a CDD claim for it.
    pub fn _register_did(
        sender: T::AccountId,
        signing_items: Vec<SigningItem>,
        protocol_fee_data: Option<(&Signatory, ProtocolOp)>,
    ) -> Result<IdentityId, DispatchError> {
        // Adding extrensic count to did nonce for some unpredictability
        // NB: this does not guarantee randomness
        let new_nonce =
            Self::multi_purpose_nonce() + u64::from(<system::Module<T>>::extrinsic_count()) + 7u64;
        // Even if this transaction fails, nonce should be increased for added unpredictability of dids
        <MultiPurposeNonce>::put(&new_nonce);

        let master_key = AccountKey::try_from(sender.encode())?;

        // 1 Check constraints.
        // 1.1. Master key is not linked to any identity.
        ensure!(
            Self::can_key_be_linked_to_did(&master_key, SignatoryType::External),
            Error::<T>::MasterKeyAlreadyLinked
        );
        // 1.2. Master key is not part of signing keys.
        ensure!(
            signing_items.iter().find(|sk| **sk == master_key).is_none(),
            Error::<T>::SigningKeysContainMasterKey
        );

        let block_hash = <system::Module<T>>::block_hash(<system::Module<T>>::block_number());
        let did = IdentityId::from(blake2_256(&(USER, block_hash, new_nonce).encode()));

        // 1.3. Make sure there's no pre-existing entry for the DID
        // This should never happen but just being defensive here
        ensure!(
            !<DidRecords>::contains_key(did),
            Error::<T>::DidAlreadyExists
        );

        // 1.4. Signing keys can be linked to the new identity.
        for s_item in &signing_items {
            if let Signatory::AccountKey(ref key) = s_item.signer {
                ensure!(
                    Self::can_key_be_linked_to_did(key, s_item.signer_type),
                    Error::<T>::AlreadyLinked
                );
            }
        }

        // 1.999. Charge the given fee.
        if let Some((payee, op)) = protocol_fee_data {
            T::ProtocolFee::charge_fee(payee, op)?;
        }

        // 2. Apply changes to our extrinsics.
        // 2.1. Link master key and add pre-authorized signing keys.
        let join_auth_data = AuthorizationData::JoinIdentity(did);
        let master_key_signatory = Signatory::from(master_key);
        Self::link_key_to_did(&master_key, SignatoryType::External, did);
        let _auth_ids = signing_items
            .iter()
            .map(|s_item| {
                Self::add_auth(
                    master_key_signatory.clone(),
                    s_item.signer.clone(),
                    join_auth_data.clone(),
                    None,
                )
            })
            .collect::<Vec<_>>();

        // 2.2. Create a new identity record.
        let record = DidRecord {
            master_key,
            ..Default::default()
        };
        <DidRecords>::insert(&did, record);

        Self::deposit_event(RawEvent::DidCreated(did, sender, signing_items));
        Ok(did)
    }

    /// It adds a new claim without any previous security check.
    fn unsafe_add_claim(
        target: IdentityId,
        claim: Claim,
        issuer: IdentityId,
        expiry: Option<T::Moment>,
    ) {
        let claim_type = claim.claim_type();
        let scope = claim.as_scope().cloned();
        let last_update_date = <pallet_timestamp::Module<T>>::get().saturated_into::<u64>();
        let issuance_date = Self::fetch_claim(target, claim_type, issuer, scope)
            .map_or(last_update_date, |id_claim| id_claim.issuance_date);

        let expiry = expiry.into_iter().map(|m| m.saturated_into::<u64>()).next();
        let pk = Claim1stKey { target, claim_type };
        let sk = Claim2ndKey { issuer, scope };
        let id_claim = IdentityClaim {
            claim_issuer: issuer,
            issuance_date,
            last_update_date,
            expiry,
            claim,
        };

        <Claims>::insert(&pk, &sk, id_claim.clone());
        Self::deposit_event(RawEvent::ClaimAdded(target, id_claim));
    }

    /// It ensures that CDD claim issuer is a valid CDD provider before add the claim.
    ///
    /// # Errors
    /// - 'UnAuthorizedCddProvider' is returned if `issuer` is not a CDD provider.
    fn unsafe_add_cdd_claim(
        target: IdentityId,
        claim: Claim,
        issuer: IdentityId,
        expiry: Option<T::Moment>,
    ) -> DispatchResult {
        let cdd_providers = T::CddServiceProviders::get_members();
        ensure!(
            cdd_providers.contains(&issuer),
            Error::<T>::UnAuthorizedCddProvider
        );

        Self::unsafe_add_claim(target, claim, issuer, expiry);
        Ok(())
    }

    pub fn is_identity_exists(did: &IdentityId) -> bool {
        <DidRecords>::contains_key(did)
    }

    /// It removes a claim from `target` which was issued by `issuer` without any security check.
    fn unsafe_revoke_claim(
        target: IdentityId,
        claim_type: ClaimType,
        issuer: IdentityId,
        scope: Option<Scope>,
    ) {
        let pk = Claim1stKey { target, claim_type };
        let sk = Claim2ndKey { scope, issuer };
        let claim = <Claims>::get(&pk, &sk);
        <Claims>::remove(&pk, &sk);
        Self::deposit_event(RawEvent::ClaimRevoked(target, claim));
    }

    /// Returns an auth id if it is present and not expired.
    pub fn get_non_expired_auth(
        target: &Signatory,
        auth_id: &u64,
    ) -> Option<Authorization<T::Moment>> {
        if !<Authorizations<T>>::contains_key(target, auth_id) {
            return None;
        }
        let auth = <Authorizations<T>>::get(target, auth_id);
        if let Some(expiry) = auth.expiry {
            let now = <pallet_timestamp::Module<T>>::get();
            if expiry > now {
                return None;
            }
        }
        Some(auth)
    }

    /// Returns identity of a signatory
    pub fn get_identity_of_signatory(signer: &Signatory) -> Option<IdentityId> {
        match signer {
            Signatory::AccountKey(key) => Self::get_identity(&key),
            Signatory::Identity(did) => Some(*did),
        }
    }

    fn leave_identity(signer: Signatory, did: IdentityId) -> DispatchResult {
        ensure!(Self::is_signer(did, &signer), Error::<T>::NotASigner);

        if let Signatory::AccountKey(key) = signer {
            Self::unlink_key_from_did(&key, did)
        }

        // Update signing keys at Identity.
        <DidRecords>::mutate(did, |record| {
            (*record).remove_signing_items(&[signer]);
        });

        Self::deposit_event(RawEvent::SignerLeft(did, signer));
        Ok(())
    }
}

impl<T: Trait> Module<T> {
    /// RPC call to know whether the given did has valid cdd claim or not
    pub fn is_identity_has_valid_cdd(
        target: IdentityId,
        leeway: Option<T::Moment>,
    ) -> Option<IdentityId> {
        Self::fetch_cdd(target, leeway.unwrap_or_default())
    }

    /// RPC call to query the given ticker did
    pub fn get_asset_did(ticker: Ticker) -> Result<IdentityId, &'static str> {
        Self::get_token_did(&ticker)
    }

    /// Retrieve DidRecords for `did`
    pub fn get_did_records(did: IdentityId) -> RpcDidRecords<AccountKey, SigningItem> {
        if <DidRecords>::contains_key(did) {
            let record = <DidRecords>::get(did);
            RpcDidRecords::Success {
                master_key: record.master_key,
                signing_items: record.signing_items,
            }
        } else {
            RpcDidRecords::IdNotFound
        }
    }

    /// Use to get the filtered link data for a given signatory
    /// - if link_type is None then return links data on the basis of the `allow_expired` boolean
    /// - if link_type is Some(value) then return filtered links on the value basis type in conjunction
    ///   with `allow_expired` boolean condition
    pub fn get_filtered_links(
        signatory: Signatory,
        allow_expired: bool,
        link_type: Option<LinkType>,
    ) -> Vec<Link<T::Moment>> {
        let now = <pallet_timestamp::Module<T>>::get();

        if let Some(type_of_link) = link_type {
            <Links<T>>::iter_prefix(signatory)
                .filter(|link| {
                    if !allow_expired {
                        if let Some(expiry) = link.expiry {
                            if expiry < now {
                                return false;
                            }
                        }
                    }
                    match link.link_data {
                        LinkData::DocumentOwned(..) => type_of_link == LinkType::DocumentOwnership,
                        LinkData::TickerOwned(..) => type_of_link == LinkType::TickerOwnership,
                        LinkData::AssetOwned(..) => type_of_link == LinkType::AssetOwnership,
                        LinkData::NoData => type_of_link == LinkType::NoData,
                    }
                })
                .collect::<Vec<Link<T::Moment>>>()
        } else {
            <Links<T>>::iter_prefix(signatory)
                .filter(|l| {
                    if !allow_expired {
                        if let Some(expiry) = l.expiry {
                            if expiry < now {
                                return false;
                            }
                        }
                    }
                    return true;
                })
                .collect::<Vec<Link<T::Moment>>>()
        }
    }

    pub fn get_did_status(dids: Vec<IdentityId>) -> Vec<DidStatus> {
        let mut result = Vec::with_capacity(dids.len());
        dids.into_iter().for_each(|did| {
            // is DID exist in the ecosystem
            if !<DidRecords>::contains_key(did) {
                result.push(DidStatus::Unknown);
            }
            // DID exist but whether it has valid cdd or not
            else if Self::fetch_cdd(did, T::Moment::zero()).is_some() {
                result.push(DidStatus::CddVerified);
            } else {
                result.push(DidStatus::Exists);
            }
        });
        result
    }

    /// It returns the list of flatten identities of the given identity.
    /// It runs recursively over all signing items.
    pub fn flatten_identities(id: IdentityId, max_depth: u8) -> Vec<IdentityId> {
        if <DidRecords>::contains_key(id) {
            let identity = <DidRecords>::get(id);

            identity
                .signing_items
                .into_iter()
                .flat_map(|si| match si.signer {
                    Signatory::Identity(sub_id) if max_depth > 0 => {
                        Self::flatten_identities(sub_id, max_depth - 1)
                    }
                    _ => vec![],
                })
                .chain([id].iter().cloned())
                .collect::<Vec<_>>()
        } else {
            vec![]
        }
    }

    /// Get the list of flatten keys fo the given identity.
    /// It runs recursively over all signing items.
    pub fn flatten_keys(id: IdentityId, max_depth: u8) -> Vec<AccountKey> {
        let sub_identities = Self::flatten_identities(id, max_depth);
        sub_identities
            .into_iter()
            .flat_map(|sub_id| {
                let identity = <DidRecords>::get(sub_id);
                identity
                    .signing_items
                    .iter()
                    .filter_map(|si| match si.signer {
                        Signatory::AccountKey(key) => Some(key),
                        _ => None,
                    })
                    .chain([identity.master_key].iter().cloned())
                    .collect::<Vec<_>>()
            })
            .collect::<Vec<_>>()
    }
}

impl<T: Trait> IdentityTrait for Module<T> {
    /// Fetches identity of a key.
    fn get_identity(key: &AccountKey) -> Option<IdentityId> {
        Self::get_identity(&key)
    }

    /// Fetches the caller's identity from the context.
    fn current_identity() -> Option<IdentityId> {
        <CurrentDid>::get()
    }

    /// Sets the caller's identity in the context.
    fn set_current_identity(id: Option<IdentityId>) {
        if let Some(id) = id {
            <CurrentDid>::put(id);
        } else {
            <CurrentDid>::kill();
        }
    }

    /// Fetches the fee payer from the context.
    fn current_payer() -> Option<Signatory> {
        <CurrentPayer>::get()
    }

    /// Sets the fee payer in the context.
    fn set_current_payer(payer: Option<Signatory>) {
        if let Some(payer) = payer {
            <CurrentPayer>::put(payer);
        } else {
            <CurrentPayer>::kill();
        }
    }

    /// Checks if the signer is authorized.
    fn is_signer_authorized(did: IdentityId, signer: &Signatory) -> bool {
        Self::is_signer_authorized(did, signer)
    }

    /// Checks if the keys is the master key of the identity.
    fn is_master_key(did: IdentityId, key: &AccountKey) -> bool {
        Self::is_master_key(did, &key)
    }

    /// Checks if the signer is authorized and has certain permissions.
    fn is_signer_authorized_with_permissions(
        did: IdentityId,
        signer: &Signatory,
        permissions: Vec<Permission>,
    ) -> bool {
        Self::is_signer_authorized_with_permissions(did, signer, permissions)
    }

    /// Adds systematic CDD claims.
    fn unsafe_add_systematic_cdd_claims(targets: &[IdentityId], issuer: SystematicIssuers) {
        targets.iter().for_each(|new_member| {
            Self::unsafe_add_claim(
                *new_member,
                Claim::CustomerDueDiligence,
                issuer.as_id(),
                None,
            )
        });
    }

    /// Removes systematic CDD claims.
    fn unsafe_revoke_systematic_cdd_claims(targets: &[IdentityId], issuer: SystematicIssuers) {
        targets.iter().for_each(|removed_member| {
            Self::unsafe_revoke_claim(
                *removed_member,
                ClaimType::CustomerDueDiligence,
                issuer.as_id(),
                None,
            )
        });
    }

    // Provides the DID status for the given DID
    fn has_valid_cdd(target_did: IdentityId) -> bool {
        Self::has_valid_cdd(target_did)
    }
}

impl<T: Trait> ChangeMembers<IdentityId> for Module<T> {
    /// Updates systematic CDDs of members of a group.
    fn change_members_sorted(
        incoming: &[IdentityId],
        outgoing: &[IdentityId],
        _new: &[IdentityId],
    ) {
        // Add/remove Systematic CDD claims for new/removed members.
        let issuer = SystematicIssuers::CDDProvider;
        Self::unsafe_add_systematic_cdd_claims(incoming, issuer);
        Self::unsafe_revoke_systematic_cdd_claims(outgoing, issuer);
    }
}

impl<T: Trait> InitializeMembers<IdentityId> for Module<T> {
    /// Initializes members of a group by adding systematic claims for them.
    fn initialize_members(members: &[IdentityId]) {
        Self::unsafe_add_systematic_cdd_claims(members, SystematicIssuers::CDDProvider);
    }
}
